__doc__="""EconoAccount - keeping track of expenses
By Peter Bengtsson

Credits to: jimbag (http://www.zope.org/Members/jimbag) for the ZCalendar"""

import string, os

from Products.PythonScripts.standard import url_quote, url_quote_plus
from Globals import MessageDialog, DTMLFile, \
                    package_home, InitializeClass
from OFS import SimpleItem, Folder
from AccessControl import ClassSecurityInfo
from DocumentTemplate import sequence
from DateTime import DateTime

import EconoZCalendar
import EconoUtils


# Product level language
LANGUAGE = "en"
DEFAULT_LANGUAGE = "en"
from Languages import EconoI18N
Lp = EconoI18N(LANGUAGE, DEFAULT_LANGUAGE)


true = TRUE = 1
false = FALSE = 0

CURRENCY_SYMBOLS = {
    'GBP':'%s',
    'USD':'$%s',
    'SEK':'%s kr'
    }

# Meta types
ACCOUNT_META_TYPE = 'Econo Account'
EXPENDITURE_META_TYPE = 'Econo Expenditure'
SPENDCODE_META_TYPE = 'Econo Spendcode'
TIMESKIP_META_TYPE = 'Econo Timeskip'


TRANSTAB_TABLE = string.maketrans('. ',
                                  '__eeouaiaOUAAEE')

manage_addEconoAccountForm = DTMLFile('econoaccountform', globals(), L=Lp)

def manage_addEconoAccount(dispatcher,id="econo",
                           title="Econo account",
                           currency='GBP', week_starts='monday',
                           language="de", amountformat='%.2f',
                           REQUEST=None):
    """ add EconoAccount object """
    dest = dispatcher.Destination()
    econoaccount = EconoAccount(id, title, currency, week_starts,
                                language=language, amountformat=amountformat)
    dest._setObject(id, econoaccount)
    self = dest._getOb(id)
    self.DeployStandards()
    L = self.getL()
    if L.whatLanguage() != language:
        L.changeLanguage(language)
    
    if REQUEST is not None:
        return MessageDialog(title=Lp['ECONO_ADDED_TITLE'],
              message=Lp['ECONO_ADDED_TEXT'], 
              action=REQUEST['URL1']+'/manage_main')


# # # # # # # # # # #
#
#  EconoAccount
#
#


class EconoAccount(Folder.Folder):
    """ Econo Account container class """
    
    meta_type=ACCOUNT_META_TYPE
    manage_options = Folder.Folder.manage_options[:-2] +\
                  ({'label': 'Management', 'action': 'management_form'},
                   )

    security=ClassSecurityInfo()
    L = EconoI18N()
  

    _properties=({'id':'title', 'type':'string', 'mode':'w'},
                 {'id':'currency',    'type': 'string',
                  'mode':'w', 'init_value':'GBP'},
                 {'id':'language',    'type': 'string',
                  'mode':'w','init_value':'en'},
                 {'id':'week_starts', 'type': 'string',
                  'mode':'w', 'init_value':'Monday'},
                 {'id':'use_users',   'type': 'boolean',
                  'mode':'w', 'init_value':0},
                 {'id':'use_advanced','type': 'boolean','mode':'w'},
                 {'id':'use_cookies', 'type': 'boolean','mode':'w'},
                 {'id':'dateformat',  'type': 'string', 'mode':'w'},
                 {'id':'amountformat','type': 'string', 'mode':'w'},
                 {'id':'switchedoff_plugins','type': 'lines', 'mode':'w'},
                 {'id':'expenditure_batchsize','type': 'int',
                  'mode':'w', 'init_val':20},
                 {'id':'methods2cache','type': 'lines',
                  'mode':'w', 'init_value':['index_html']}
                 )
    
    statistics_methods = ({'id':'basic_statistics',
                           'title':L['BASIC_STATISTICS_TITLE'],
                           'description':L['BASIC_STATISTICS_DESCRIPTION'],
                           'author':'Peter Bengtsson'},)
    backup_methods = ()
    input_methods = ()

    deploy_standards_methods = ()

    def __init__(self, id, title, currency, week_starts,
                 use_users=false, use_advanced=false,
                 dateformat='%Y-%m-%d', amountformat='%.2f',
                 language=LANGUAGE):
        self.id = id
        self.title = title
        self.currency = currency
        self.week_starts = week_starts
        self.use_users = false
        self.use_advanced = false
        self.use_cookies = 0
        self.spendcode_order=[]
        self.userid_order=[]
        self.dateformat = dateformat
        self.amountformat = amountformat
        self.last_merge = None
        self.switchedoff_plugins = []
        self.expenditure_batchsize = 40
        self.methods2cache =['index_html']
        self.language = language

    def getL(self):
        """ return EconoI18N object """
        l = EconoAccount.L
        # if product has been refreshed
        if l.whatLanguage() != self.language:
            l.changeLanguage(self.language)
        return l
    
    security.declareProtected('Add Documents, Images, and Files',
                              'DeployStandards')
    def DeployStandards(self, REQUEST=None, return_touched=0):
        """ takes some files and creates Zope objects in instance """
        touched = {}
        L = self.getL()
        osj = os.path.join
        producthome = osj(package_home(globals()))
        standards_home = osj(producthome,'standards')
        root = self.getRoot()

        rootfolders = {'expenditures':L['EXPENDITURES_WORD'],
                       'spendcodes':L['SPENDCODES_WORD'],
                       'users':L['USERS_WORD'],
                       'timeskips':L['TIMESKIPS_WORD']
                       }
        for folder, title in rootfolders.items():
            if not hasattr(root, folder):
                root.manage_addFolder(folder, title=title)
                touched[folder]='Folder'
        if not hasattr(root, 'calendar'):
            EconoZCalendar.manage_addCalendar(root, 'calendar',
                                         week_starts=self.week_starts)
            touched['calendar']='EconoZCalendar'
        if not hasattr(root, 'econoaccountlogo.gif'):
            touched['econoaccountlogo.gif']='Image'            
            o = osj(producthome,'www','econoaccountlogo.gif')                              
            root.manage_addImage('econoaccountlogo.gif',\
                                 file=open(o,'rb').read(), title=root.title)

        for filestr in os.listdir(standards_home):
            if filestr[-5:] == '.dtml':
                id, title = EconoUtils.cookIdAndTitle(filestr.replace('.dtml',''))
                if not root._hasattr_itself(id):
                    touched[id]='DTML Method'
                    file = DTMLFile('standards/%s'%filestr.replace('.dtml',''), globals()).read()
                    root.manage_addDTMLMethod(id, title, file=file)

        for method in self.getDeployStandardsMethods():
            if type(method)==type('s'):
                touched = getattr(self, method)(touched=touched)
            else:
                touched = getattr(self, method['id'])(touched=touched)

        if return_touched:
            return touched
        elif REQUEST is None:
            return L['STANDARDS_DEPLOYED_TIME'] + " %s "% DateTime()
        else:
            s=''
            for id, meta_type in touched.items():
                s= s+'%s: %s<br>'%(meta_type, id)
            if not touched:
                s= "<em>%s</em>"%L['STANDARDS_NONE_NEW']
            s='<table><tr><td>%s</td></tr></table>'%s
            if self.REQUEST.has_key('HTTP_REFERER') and \
               self.REQUEST.HTTP_REFERER != '':
                returnurl = REQUEST['URL1']+'/management_form'
            else:
                returnurl = REQUEST['URL1']
            return MessageDialog(title=L['STANDARDS_ADDED_TITLE'],
              message='%s<br><br><br>%s'%(L['STANDARDS_ADDED_TEXT'],s), 
              action=returnurl)
                    
    def _hasattr_itself(object,attribute):
        """return true iff *object* itself has attribute *attribute*."""
        base= getattr(object,'aq_base',object)
        return hasattr(base,attribute)

    def getDateformat(self):
        """ return current date format """
        if not hasattr(self, 'dateformat'):
            self.dateformat = '%Y-%m-%d'
        return self.dateformat

    def getAmountformat(self):
        """ return amount number format """
        if not hasattr(self, 'amountformat'):
            self.amountformat = '%.2f'
        return self.amountformat
        
    security.declareProtected('Access contents information', 'getRoot')
    def getRoot(self):
        """ Get the root instance object """
        if self.meta_type == ACCOUNT_META_TYPE:
            return self
        else:
            for sp in self.superValues(ACCOUNT_META_TYPE):
                if sp.id in self.REQUEST.URL1.split('/'):
                    return sp

    security.declareProtected('Access contents information', 'getRootURL')
    def getRootURL(self):
        """ the the absolute url of the root instance """
        return self.getRoot().absolute_url()

    # alias
    getRootUrl = getRootURL

    security.declareProtected('View','index_html') 
    def index_html(self, *arg, **kw):
        " Zope __str__  "
        return DTMLFile('index', globals(), mapping=kw).__of__(self)()


    security.declareProtected('View', 'getCurrency')
    def getCurrency(self):
        " return currency for display "
        return self.currency

    security.declareProtected('View', 'HeadingLinks')
    def HeadingLinks(self, display, sortname, default=0, startreversed=0):
        """Returns a hyperlink that can
           be used for resorting the listing
        """
        L = self.getL()
        request = self.REQUEST
        querystring = request.QUERY_STRING

        
        imgtag = """<img src="/misc_/EconoAccount/%s" 
                       width="11" height="12" hspace="2" alt="%s">"""
        ascending = imgtag%('ascarrow.gif','Ascending')
        descending = imgtag%('descarrow.gif','Descending')
        invisible = imgtag%('invisiblearrow.gif','')

        # get clean URL
        URL = self.ActionURL(request.URL)
        
        n_querystring = []
        for query in string.split(querystring,'&'):
            if string.find(query,'sortorder') == -1 and \
               string.find(query,'reverse') == -1:
                n_querystring.append(query)
        querystring = '&'.join(n_querystring)

        if querystring != '':
            querystring = '&%s'%querystring
        
        if request.has_key('reverse') and \
           (request['reverse'] == 1 or request['reverse'] == '1'):
            reverse = 1
        elif startreversed and not \
             (request.has_key('sortorder') and request['sortorder']==sortname):
            reverse = 1
        else:
            reverse = 0
        if request.has_key('sortorder') and \
           request['sortorder'] == sortname or \
           (default and not request.has_key('sortorder')):
            # have sorted by this, just let them reverse
            if reverse:
                return '<a class="headinglink_current" href="%s?sortorder=%s%s" title="%s %s">%s</a>%s' \
                       % (URL, sortname, querystring,
                          L['SORT_BY'], display, display, descending)
            else:
                return '<a class="headinglink_current" href="%s?sortorder=%s&reverse=1%s" title="%s">%s</a>%s'\
                       % (URL, sortname, querystring,
                          L['SORT_REVERSE'], display, ascending )
        else:
            if startreversed:
                return '<a class="headinglink" href="%s?sortorder=%s%s&reverse=1" title="%s %s">%s</a>%s' \
                   % (URL, sortname, querystring,
                      L['SORT_BY'], display, display, invisible)
            else:
                return '<a class="headinglink" href="%s?sortorder=%s%s" title="%s %s">%s</a>%s' \
                   % (URL, sortname, querystring,
                      L['SORT_BY'], display, display, invisible)

    def FixMissingProperties(self, REQUEST=None):
        """ make sure all objects have all properties """
        account_properties = EconoAccount._properties + \
                             ({'id':'spendcode_order',
                               'type':'list',
                               'value':[]},
                              {'id':'userid_order',
                               'type':'list',
                               'value':[]}
                              )

        spendcode_properties = EconoSpendcode._properties
        expenditure_properties = EconoExpenditure._properties
        user_properties = EconoUser._properties

        self._apply_init_properties(self.getRoot(), account_properties)
        for spendcode in self.getSpendcodes():
            self._apply_init_properties(spendcode, spendcode_properties)
        for expenditure in self.getExpenditures():
            self._apply_init_properties(expenditure, expenditure_properties)
        for user in self.getUsers():
            self._apply_init_properties(user, user_properties)

        L = self.getL()        
        if REQUEST is not None:
            tabs_message = L['FIXED_ALL_PROPERTIES']
            return self.management_form(manage_tabs_message=tabs_message)

        
    def _apply_init_properties(self, root, properties):
        """ create properties from 'properties' in 'root' """
        for prop in properties:
            if not hasattr(root, prop['id']):
                if prop.has_key('value'):
                    init_val = prop['value']
                elif prop.has_key('init_value'):
                    init_val = prop['init_value']
                else:
                    init_val = self._getPropertyInitvalue(prop['type'])
                if type(init_val)==type('s') and \
                   not (init_val=="''" or init_val=='""'):
                    exec_cmd = "root.%s = '%s'"% (prop['id'], init_val)
                else:
                    exec_cmd = "root.%s = %s"% (prop['id'], init_val)
                exec exec_cmd
        

    def _getPropertyInitvalue(self, type):
        """ get the initial guess value depending on what type is """
        type = type.lower().strip()
        if type == 'int' or type == 'long':
            return 0
        elif type == 'boolean':
            return false
        elif type == 'list' or type == 'lines' or type == 'tokens':
            return []
        elif type == 'date':
            return DateTime()
        elif type == 'float':
            return 0.0
        else:
            return "''"
        
    def MergeExpenditures(self, REQUEST=None):
        """ Sum up all expenditures with same date and spendcode into one """
        L = self.getL()
        
        # datewise
        if self.UseAdvanced() or self.UseUsers():
            raise "CantMergeExpenditures", L['CANT_MERGE_EXPENDITURES_USERS']

        expenditures = sequence.sort(self.getExpenditures(), (('datum',),))
        before_len = len(expenditures)
        datewise = {}
        for expenditure in expenditures:
            if datewise.has_key(expenditure.datum):
                datewise[expenditure.datum].append(expenditure)
            else:
                datewise[expenditure.datum] = [expenditure]
        
        for expenditures in datewise.values():
            self._merge_expendituresums(expenditures)
            
        self.LastMerge(set=1)
        after_len = len(self.getExpenditures())
        if REQUEST is not None:
            tabs_message = L['MERGED_INTO']%(before_len, after_len)
            return self.management_form(manage_tabs_message=tabs_message)
        else:
            return before_len, after_len

    def LastMerge(self, set=0):
        """ Use to either set the latest merge date or display """
        root = self.getRoot()
        if set:
            root.last_merge = DateTime()
        else:
            if hasattr(root, 'last_merge'):
                return root.last_merge
            else:
                return None
                
    def _merge_expendituresums(self, expenditures):
        """ expect 'expenditures' to be of same datum """
        spendcode_sums={}
        can_del_ids=[]
        # Stuff spendcode_sums with 'spendcodeX':<firstobject>'
        # And then change <firstobject>'s amount for each similar
        # spendcode.
        for expenditure in expenditures:
            if spendcode_sums.has_key(expenditure.spendcode):
                first_amount = spendcode_sums[expenditure.spendcode].amount
                first_desc = spendcode_sums[expenditure.spendcode].description

                this_amount = expenditure.amount
                this_desc = expenditure.description
                
                # change the first object
                spendcode_sums[expenditure.spendcode].editExpenditure(amount=\
                                        first_amount + this_amount)
                if this_desc != '':
                    spendcode_sums[expenditure.spendcode].editExpenditure(\
                        description = first_desc +", "+ this_desc)
                can_del_ids.append(expenditure.getId())
                    
            else:
                spendcode_sums[expenditure.spendcode]=expenditure

        self.expenditures.manage_delObjects(can_del_ids)

    def DoCache(self):
        """ check if this visited page is supposed to be cached """
        method = self.REQUEST.URL.split('/')[-1]
        return method in self.methods2cache
            

    def ActionURL(self, u=None):
        """ if URL ends in 'index_html' remove that """
        if u is None:
            u = self.REQUEST.URL
        if u[-11:]=='/index_html':
            return u[:-11]
        else:
            return u

    def UseAdvanced(self):
        """ Determine if in advanced mode or not """
        return self.use_advanced

    def getExpenditureBatchSize(self, default=0):
        """ Return instance attribute """
        if self.REQUEST.has_key('bsize') and not default:
            return int(self.REQUEST.bsize)
        else:
            return self.expenditure_batchsize

    def getSwitchedOffPlugins(self):
        """ Return instance attribute method """
        return self.switchedoff_plugins

    def _filterPlugins(self, methods):
        """ filter through on avoid list """
        use_methods=[]
        switchedoff_plugins = self.getSwitchedOffPlugins()
        for method in methods:
            if method['id'] not in switchedoff_plugins:
                use_methods.append(method)
        return use_methods
    
    def getBackupMethods(self, nofilter=0):
        """ return class variable """
        methods = EconoAccount.backup_methods
        if nofilter:
            return methods
        else:
            return self._filterPlugins(methods)

    def getInputMethods(self, nofilter=0):
        """ return class variable """
        methods = EconoAccount.input_methods
        if nofilter:
            return methods
        else:
            return self._filterPlugins(methods)

    def getDeployStandardsMethods(self, nofilter=0):
        """ return class variable """
        methods = EconoAccount.deploy_standards_methods
        if nofilter:
            return methods
        else:
            return self._filterPlugins(methods)
    
    def getStatisticsMethods(self, nofilter=0):
        """ return class variable """
        methods = EconoAccount.statistics_methods
        if nofilter:
            return methods
        else:
            return self._filterPlugins(methods)

    def DisplayAmount(self, amount):
        """ Return amount with right display format """
        return self.getAmountformat()%amount
    
    def CurrencySymbol(self, s, append_failed_text=0):
        """ combine currency symbol with amount if possible """
        if type(s)!=type('string'):
            s = self.DisplayAmount(s)
            
        if CURRENCY_SYMBOLS.has_key(self.currency.upper()):
            s = CURRENCY_SYMBOLS[self.currency.upper()]%s
        elif append_failed_text:
            s = s + " " + self.currency.upper()
        return s
        
                        
    # Templates
    security.declareProtected('View', 'expenditure_fulllist')
    expenditure_fulllist = DTMLFile('expenditure_fulllist', globals())
    
    security.declareProtected('View', 'add_expense_form')
    add_expense_form = DTMLFile('add_expense_form', globals())

    security.declareProtected('View', 'services_index')
    services_index = DTMLFile('services_index', globals())

    security.declareProtected('View', 'expendituresListTable')
    expendituresListTable = DTMLFile('expendituresListTable', globals())

    security.declareProtected('View', 'list_expenses')
    list_expenses = DTMLFile('list_expenses', globals())

    security.declareProtected('View', 'manage_spendcodes')
    manage_spendcodes = DTMLFile('manage_spendcodes', globals())

    security.declareProtected('Access contents information', 'explain_error')
    explain_error = DTMLFile('explain_error', globals())

    process_placeholder = DTMLFile('process_placeholder', globals())
    management_form = DTMLFile('management_form', globals())

    # ---------- TIMESKIPS STUFF ---------------------------------

    def TotalTimeskipDuration(self):
        """ Return total amount of days all timeskips add up to """
        all_timeskips = self.getTimeskips()
        sum = 0
        for timeskip in all_timeskips:
            sum += timeskip.getDuration("int")
        return sum
        
    def DatesAlreadyUsed(self, dates=[]):
        """ Check if a date already has been used for an expenditure.
            This method can take some long time."""
        expenditures = self.getExpenditures()

        if type(dates) != type([]):
            dates = [dates]
        dates = self._dates2datestrings(dates)
        count = 0
        dateformat = self.getDateformat()
        for expenditure in expenditures:
            if expenditure.datum.strftime(dateformat) in dates:
                count += 1
        return count

    def _dates2datestrings(self, dates=[]):
        """ convert all dates as strings in a list """
        n_dates = []
        dateformat = self.getDateformat()
        for date in dates:
            n_dates.append(date.strftime(dateformat))
        return n_dates
        
        
    def DateAlreadySkipped(self, date):
        """ Check if a data is already used in one timskip.
            For dates exactly on start or end of a timeskip return false"""
        if type(date) == type('s'):
            # Convert as date object
            date = DateTime(date)
        for timeskip in self.timeskips.objectValues(TIMESKIP_META_TYPE):
            if timeskip.DateSpanned(date):
                return true
        return false

    def _getTimeskips(self, only_id=0):
        """ crudely get the timeskips objects """
        if only_id:
            return self.timeskips.objectIds(TIMESKIP_META_TYPE)
        else:
            return self.timeskips.objectValues(TIMESKIP_META_TYPE)
        
    security.declareProtected('View', 'getTimeskips')
    def getTimeskips(self):
        """ Return timeskipobjects and possibly do some special sorting """
        request = self.REQUEST
        
        timeskips = self._getTimeskips()
        if request.has_key('sortorder') and request['sortorder']:
            timeskips_byduration = self._filterByDuration(timeskips)
            try:
                if request.has_key('reverse') and int(request['reverse']):
                    timeskips_byduration.reverse()
                return timeskips_byduration
            except:
                return timeskips_byduration
        else:
            return timeskips

    def _filterByDuration(self, objects):
        """ Get their durations and sort by that """
        n_objects=[]
        object_durations={}
        for object in objects:
            object_durations[object.getDuration()] = object
        durations = object_durations.keys()
        durations.sort()
        #durations.reverse()
        for each in durations:
            n_objects.append(object_durations[each])
        return n_objects

    security.declareProtected('Delete objects', 'manage_delTimeskips')
    def manage_delTimeskips(self, ids=[], REQUEST=None):
        """ delete timeskip objects """
        L = self.getL()
        if type(ids) == type('s'):
            ids = [ids]
        self.timeskips.manage_delObjects(ids)
        
        if REQUEST is not None:
            tabs_message = L['TIMESKIPS_DELETED']
            url = self.getRootURL()+'/manage_timeskips'
            url = '%s?manage_tabs_message=%s'%(url, tabs_message)
            self.REQUEST.RESPONSE.redirect(url)


    security.declareProtected('View', 'manage_addTimeskip')
    def manage_addTimeskip(self, start_date=None, end_date=None,
                           description=''):
        """ create timeskip object """
        L = self.getL()
        request = self.REQUEST
        
        SubmitError={}
        start_date_test = self._IsValidDateInput(start_date)
        end_date_test = self._IsValidDateInput(end_date)
        if start_date_test and end_date_test:
            # Check that start_date really comes before end_date
            start_date = str(start_date).replace('-','/')
            end_date = str(end_date).replace('-','/')
            if DateTime(end_date)>DateTime(start_date):

                # Check that there are no expenditures during this timeskip
                day_dates=[]
                difference = int(DateTime(end_date)- DateTime(start_date))
                for day in range(1, difference):
                    day_dates.append(DateTime(start_date)+day)
                    # day_dates takes only date values _between_ the
                    # start_date and end_date.

                dates_already_used = self.DatesAlreadyUsed(day_dates)
                if dates_already_used:
                    SubmitError[L['DATE_CAPTION']] = \
                           L['EXPENDITURE_DURING_TIMESKIP']%dates_already_used
                else:

                    # nothing left to check
                    dates = self._purify_date(str(start_date)+str(end_date))
                    id = self._generateTimestampedId(dates)
                    timeskip = EconoTimeskip(id, start_date, end_date, description)
                    timeskip_object = self.timeskips._setObject(id, timeskip)

                    url = '%s?manage_tabs_message=%s&newtimeskipid=%s'%\
                      (self.getRootURL()+'/manage_timeskips',
                       url_quote_plus(L['TIMESKIP_ADDED']),
                       id)
                    request.RESPONSE.redirect(url)
            else:
                SubmitError[L['DATE_CAPTION']] = L['ENDDATE_BEFORE_STARTDATE_ERROR']
        else:
            if not start_date_test:
                SubmitError[L['FROMDATE_CAPTION']] = L['FROMDATE_INVALID']
            if not end_date_test:
                SubmitError[L['TODATE_CAPTION']] = L['TODATE_INVALID']

        if SubmitError:
            return self.manage_timeskips(SubmitError=SubmitError)

##    def zfillDateString(self, dstr):
##        """ Zero fill date string.
##            Effectivly convert 2002-7-22 to 2002-07-22
##        """
##        
##        if len(dstr.split('/'))==3:
##            splitter = '/'
##        elif len(dstr.split('-'))==3:
##            splitter = '-'
##
##        splitted = dstr.split(splitter)
##        for c in range(len(splitted)):
##            if len(splitted[c])==1:
##                splitted[c]=string.zfill(splitted[c],2)
##        return splitter.join(splitted)
    
    def _purify_date(self, d, p=''):
        return d.replace('-','').replace('/','').replace(' ',p)


    security.declareProtected('Delete objects', 'manage_removeTimeskips')
    def manage_removeTimeskips(self, ids=[], DestinationURL=None,
                                 redirect=1):
        """ simply remove some objects """
        request = self.REQUEST
        L = self.getL()
        if type(ids)==type('s'):
            ids=[ids]
        self.timeskips.manage_delObjects(ids)
        if redirect:
            if DestinationURL:
                request.RESPONSE.redirect(DestinationURL)
            else:
                url = "%s?manage_tabs_message=%s&r=%s"%\
                      (self.getRoot().absolute_url(),
                       url_quote_plus(L['TIMESKIPS_REMOVED']),
                       int(DateTime()))
                request.RESPONSE.redirect(url)
        

    security.declareProtected('View', 'manage_timeskips')
    manage_timeskips = DTMLFile('manage_timeskips', globals())

    security.declareProtected('View', 'calendar_window')
    calendar_window = DTMLFile('calendar_window', globals())


    # ---------- EXPENDITURE STUFF -------------------------------

    def _safeTitle2Id(self, title):
        " from a title return a safe id "
        L = self.getL()
        title = title.strip()
        title = string.translate(title, TRANSTAB_TABLE)
        id = EconoUtils.safeId(title)
        if id=='':
            raise "NoId", L['TITLE_NO_ALLOWED_CHARS']
        return id


    security.declareProtected('Add Econo Accounts','manage_addExpenditure')
    def manage_addExpenditure(self):
        """ wrap around addExpenditure() but control what's in REQUEST
            with my own error message """
        L = self.getL()
        request = self.REQUEST
        SubmitError={}
        params_needed = {'amount':L['AMOUNT_CAPTION'],
                         'spendcode':L['SPENDCODE_CAPTION'],
                         'datum':L['DATE_CAPTION']}
        if self.UseUsers():
            params_needed['userid'] =L['USER_CAPTION']
        for param, name in params_needed.items():
            if not request.has_key(param) or \
               (request.has_key(param) and request.get(param)==''):
                SubmitError[name] = L['SUBMITERROR_NOINPUT']
            # amount was submitted, now check it
            elif param=='amount':
                try:
                    a=self._tuneAmount(request.get('amount'))
                except:
                    SubmitError[L['AMOUNT_CAPTION']]= L['INVALID_WORD']
            elif param=='datum':
                if not self._IsValidDateInput(request.get('datum')):
                    SubmitError[L['DATE_CAPTION']] = L['INVALID_WORD']
                elif self.DateAlreadySkipped(request.get('datum')):
                    SubmitError[L['DATE_CAPTION']] = L['DATE_ALREADY_SKIPPED']
            elif param=='spendcode':
                if not self._spendcodeExists(request.get('spendcode').strip()):
                    SubmitError[L['SPENDCODE_CAPTION']]=\
                                                L['SUBMITERROR_NO_SPENDCODE']
            elif param=='userid':
                if not self.userExists(request.get('userid').strip()):
                    SubmitError[L['USER_CAPTION']] = L['SUBMITERROR_NO_USER']
            

        if SubmitError:
            return self.index_html(SubmitError=SubmitError)
        else:
            newid = self.addExpenditure(request.amount,
                                request.spendcode,
                                request.datum,
                                request.get('userid','').strip(),
                                request, return_id=1)
            # to prevent silly reloads<->re-add
            if request.has_key('sortorder'):
                sortorder = '&sortorder=%s'%request.get('sortorder')
            else:
                sortorder = ''
            if request.has_key('reverse'):
                reverse = '&reverse=%s'%request.get('reverse')
            else:
                reverse = ''
            if request.has_key('CalendarMonth'):
                calendarmonth = '&CalendarMonth=%s'%request.get('CalendarMonth')
            else:
                calendarmonth = ''
            url = '%s?manage_tabs_message=%s&datum=%s&spendcode=%s&userid=%s&newexpid=%s%s%s%s'%\
                  (self.getRootURL(),
                   url_quote_plus(L['EXPENDITURE_ADDED']),
                   request.datum,
                   request.spendcode,
                   request.get('userid','').strip(),
                   newid,
                   sortorder,
                   reverse,
                   calendarmonth)
            request.RESPONSE.redirect(url)
                    

    def _IsValidDateInput(self, d):
        """ Check if a supposed to be date string is valid """
        passed_date_test=0
        if self._IsValidDateString(d):
            try:
                datum = DateTime(d)
                passed_date_test=1
            except:
                pass
        return passed_date_test
    
    def _IsValidDateString(self, d):
        """ do a couple of checks on a date string.
            Just wrapping a DateTime('date') with try: seems not to work."""
        d=d.replace('-','').replace('/','').replace(' ','').replace(':','')
        d=d.replace('@','')
        try:
            d=int(d)
            return true
        except:
            return false
        
    def _spendcodeExists(self, spendcode):
        """ check if a spendcode exists """
        spendcode_ids = self._getSpendcodeIds()
        if not spendcode in self._getSpendcodeIds():
            return false
        else:
            return true

    security.declareProtected('Add Econo Accounts','addExpenditure')                       
    def addExpenditure(self, amount, spendcode, datum, userid='',
                       REQUEST=None, return_id=0):
        """ add the expenditure object """
        L = self.getL()
        
        # check data
        amount = self._tuneAmount(amount)
        spendcode = str(spendcode).strip()
        if not self._spendcodeExists(spendcode):
            raise "NoSpendcode", L['SPENDCODE_NOT_EXIST']%spendcode

        if type(datum)==type('s'):
            datum=DateTime(datum)

        id = self._generateTimestampedId(amount)
        self._addExpenditure(id, amount, datum, spendcode, userid)

        self._orderSpendcodes()

        if return_id:
            return id
        elif REQUEST is not None:
            tabs_message = L['EXPENDITURE_ADDED']
            return self.index_html(manage_tabs_message=tabs_message)
        else:
            return id

    def _addExpenditure(self, id, amount, datum, spendcode, userid):
        """ create object """
        exp_cont = self.expenditures
        expenditure = EconoExpenditure(id, amount, datum, spendcode, userid)
        exp_cont._setObject(id, expenditure)
        return getattr(exp_cont, id)

    def _tuneAmount(self, a):
        """ Fix amount to try to make it a float.
            Let the amount be a mathematical expression recognizing
            +, -, *, / and ()"""
        L = self.getL()

        operators = '+-*/'
        recognize = '0123456789.+-*/()'

        # prepare amount string
        a=str(a).strip().replace(',','.').replace(' ','')
        
        # make sure only 'recognize' are in 'a'
        for e in list(a):
            if not e in list(recognize):
                raise "AmountNotFloat", L['AMOUNT_NOT_FLOAT']
            
        # disallow operators in start or end
        for operator in operators:
            if a.startswith(operator) or a.endswith(operator):
                raise "AmountNotFloat", L['OPERATOR_NOT_INSIDE']

        # evaluate amount now
        try:
            exec "a=%s"%a
        except:
            raise "AmountNotFloat", L['AMOUNT_NOT_FLOAT']

        # return float point number
        return float(a)
               
    def _generateTimestampedId(self, endbit=None, randomness=1):
        " return fairly random string based on timestamp and amount "
        timestamp = str(int(DateTime()))
        id = timestamp
        if endbit:
            id = id+'_%s'%str(endbit)
        if randomness and randomness >= 1:
            id = id+'_%s'% EconoUtils.generateRandomMessageId(3*randomness)
        return id

    def _getExpenditures(self, only_id=0):
        " crudely pick up the objects "
        if only_id:
            return self.expenditures.objectIds(EXPENDITURE_META_TYPE)
        else:
            return self.expenditures.objectValues(EXPENDITURE_META_TYPE)

    security.declareProtected('View', 'getExpenditures')
    def getExpenditures(self, spendcodes=[], userids=[]):
        """ return all expenditure objects """
        all_expenditures = self._getExpenditures()
        if spendcodes and len(spendcodes)>0:
            new_all_expenditures=[]
            for expenditure in all_expenditures:
                if expenditure.spendcode in spendcodes:
                    new_all_expenditures.append(expenditure)
            a = new_all_expenditures
        else:
            a = all_expenditures

        if userids and len(userids)>0:
            new_a = []
            for expenditure in a:
                if expenditure.userid in userids:
                    new_a.append(expenditure)
            a = new_a
        return a

    security.declareProtected('Delete objects', 'manage_removeExpenditure')
    def manage_removeExpenditure(self, ids=[], DestinationURL=None,
                                 redirect=1):
        """ simply remove some objects """
        if type(ids)==type('s'):
            ids=[ids]
        self.expenditures.manage_delObjects(ids)
        if redirect:
            if DestinationURL:
                self.REQUEST.RESPONSE.redirect(DestinationURL)
            else:
                url = "%s?manage_tabs_message=%s&r=%s"%\
                      (self.getRootURL(),
                       url_quote_plus("Expenditure removed."),
                       int(DateTime()))
                self.REQUEST.RESPONSE.redirect(url)

    def ExpenditureBatchLinks(self, total_size, batch_size, next_number):
        """ return a string of html of batch links """
        s=''
        qs = self.ExpenditureBatchLinksQueryString(self.REQUEST)
        links=[]
        link_count=0
        if qs!='':
            qs="&%s"%qs
        
        for x in range(1,total_size, batch_size):
            x0 = x
            x1 = x+batch_size-1
            if next_number-1 == x1:
                #s += "%s - %s"%(x0, x1)
                if total_size-x0 < x1-x0:
                    x1 = total_size
                
                links.append("%s - %s"%(x0, x1))
                focus_position = link_count
            else:
                if total_size - x0 < x1-x0:
                    x1 = total_size
                #s += '<a href="?query_start=%s%s">%s - %s</a>'%(x0, qs, x0, x1)
                links.append('<a href="?query_start=%s%s">%s - %s</a>'%(x0, qs, x0, x1))
                link_count += 1


            s += " - "
        if link_count == focus_position:
            # far right
            to_left = 2
            to_right = 1
            extra = -1
        elif focus_position == 0:
            # far left
            to_left = -2
            to_right = 1
            extra = -1
        else:
            to_left = -1
            to_right = 1
            extra = 0
        if len(links) > 5:
            for c in range(len(links)):
                if focus_position not in [c+to_left,c,c+to_right,c+extra]:
                    links[c] = '...'
        return " - ".join(self._removeDuplicateDots(links,'...'))

    def _removeDuplicateDots(self, links, dotstr='...'):
        """ Make ['a','...','...','r','s','...']
            Into ['a','...','r','s','...']
        """
        n_links=[]
        current=None
        for link in links:
            if not (link == current and link ==dotstr):
                n_links.append(link)
            current=link
        return n_links
                

    def ExpenditureBatchLinksQueryString(self, request):
        """ take out all the stuff from the request needed in batch links """
        qs=[]
        for each in ['sortorder','reverse','bsize']:
            if request.has_key(each):
                qs.append('%s=%s'%(each, url_quote_plus(request.get(each))))
        return '&'.join(qs)

            

    # ---------- SPENDCODE STUFF ---------------------------------

    security.declareProtected('View management screens', 'addSpendcode')
    def addSpendcode(self, title, description='', REQUEST=None):
        """ add spendcode """
        L = self.getL()
        all_spendcodes = self.spendcodes.objectValues(SPENDCODE_META_TYPE)
        for spendcode in all_spendcodes:
            if spendcode.title.strip().lower() == title.strip().lower():
                raise "SpendcodeTaken", L['SPENDCODE_TAKEN']
                return
        id = self._safeTitle2Id(title).lower()
        spendcode = EconoSpendcode(id, title, description)
        self.spendcodes._setObject(id, spendcode)

        if REQUEST is not None:
            tabs_message = L['SPENDCODE_ADDED']
            return self.manage_spendcodes(self.REQUEST, 
                                          manage_tabs_message=tabs_message)

    security.declareProtected('View', 'getSpendcodeFormSize')
    def getSpendcodeFormSize(self, nr_spendcodes):
        """ for the form """
        nr = nr_spendcodes
        if nr <= 15:
            return nr
        elif nr > 15:
            return 16

    security.declareProtected('Access contents information', 'getSpendcode')
    def getSpendcode(self, id):
        """from id return object """
        self.getL()
        if hasattr(self.spendcodes, id):
            return getattr(self.spendcodes, id)
        else:
            raise "InexistingSpendcode", L['SPENDCODE_BYID_NOT_EXIST']
            
    def _getSpendcodeIds(self):
        """ wrapper around _getSpendcodes(only_id=1) """
        return self._getSpendcodes(only_id=1)

    def _getSpendcodes(self, only_id=0):
        " crudely pick up the objects "
        if only_id:
            return self.spendcodes.objectIds(SPENDCODE_META_TYPE)
        else:
            return self.spendcodes.objectValues(SPENDCODE_META_TYPE)

    security.declareProtected('View', 'getSpendcodes')
    def getSpendcodes(self, sort=0, ids=[]):
        """ return all spendcodes, possibly sorted """
        # objects
        
        if ids:
            # a list of ids of spendcodes is provided.
            # get them down as objects.
            os = []
            for id in ids:
                os.append(getattr(self.spendcodes, id))
            sort = 0
        else:
            os = self._getSpendcodes()

        if os is None or len(os)==0:
            return None
        elif sort and self.spendcode_order:
            oss=[]
            for each in self.spendcode_order:
                oss.append(getattr(self.spendcodes,each))
            for each in os:
                if each not in oss:
                    oss.append(each)
            
            return oss
        else:
            return os

    def _orderSpendcodes(self):
        """ look at all spendcodes and list then in order
            where 1st one is the spendcode most frequently used.
            Do this to change the 'spendcode_order' list property. """
        frequencies = {}
        for expenditure in self.expenditures.objectValues(EXPENDITURE_META_TYPE):
            spendcode = expenditure.spendcode
            if frequencies.has_key(spendcode):
                frequencies[spendcode] = 1+ frequencies[spendcode]
            else:
                frequencies[spendcode] = 1
        numbers = frequencies.values()
        numbers.sort()
        numbers.reverse()
        spendcodes=[]
        while frequencies:
            numbers = frequencies.values()
            numbers.sort()
            numbers.reverse()
            for number in numbers:
                frequencies, spendcode = self._popDictOnValue(frequencies,
                                                              number)
                spendcodes.append(spendcode)
        self.spendcode_order = spendcodes


    def _popDictOnValue(self, dict, val):
        """ return the key from a dict whose value is given,
            then delete that item """
        the_key=''
        for key, value in dict.items():
            if value == val:
                the_key = key
                break
        del dict[key]
        return dict, the_key
    
    security.declareProtected('Delete objects', 'manage_removeSpendcode')
    def manage_removeSpendcode(self, ids=[], DestinationURL=None):
        """ simply remove some objects """
        L = self.getL()
        if type(ids)==type('s'):
            ids=[ids]

        # Remove all expenditures using this spendcode
        for id in ids:
            self._remove_expenditures_by_spendcode(id)

        self.spendcodes.manage_delObjects(ids)

        # Make sure there are no loose references to this spendcode
        self._orderSpendcodes()

        if DestinationURL:
            self.REQUEST.RESPONSE.redirect(DestinationURL)
        else:
            url = "%s?manage_tabs_message=%s&r=%s"%\
                  (self.getRoot().absolute_url()+'/manage_spendcodes',
                   url_quote_plus(L['SPENDCODE_REMOVED']),
                   int(DateTime()))
            self.REQUEST.RESPONSE.redirect(url)

    def _remove_expenditures_by_spendcode(self, spendcodeid):
        """ remove all expendtures who use this spendcode """
        all_expenditures = self._getExpenditures()
        remove_expenditures = []
        for expenditure in all_expenditures:
            if expenditure.spendcode == spendcodeid:
                remove_expenditures.append(expenditure.getId())
        self.expenditures.manage_delObjects(remove_expenditures)

        
    # ---------- FOR BASIC_STATISTICS -----------------------------

    # Statistics templates
    security.declareProtected('View','basic_statistics')
    basic_statistics = DTMLFile('basic_statistics', globals())

    security.declareProtected('View', 'StartDate')
    def StartDate(self):
        """ return the oldest expenditures 'datum' """
        # get ALL expenditures
        expenditures = self._getExpenditures()
        # sort them by datum attribute
        expenditures = sequence.sort(expenditures, (('datum',),))
        # return the date of the oldest
        if len(expenditures) > 0:
            return expenditures[0].datum
        else:
            return DateTime()

    security.declareProtected('View', 'EndDate')
    def EndDate(self):
        """ return the least oldest expenditures 'datum' """
        # get ALL expenditures
        expenditures = self._getExpenditures()
        # sort them by datum attribute
        expenditures = sequence.sort(expenditures, (('datum',),))
        expenditures.reverse()
        # return the date of the oldest
        if len(expenditures) > 0:
            return expenditures[0].datum
        else:
            return DateTime()

    security.declareProtected('View', 'StrDateDifference')
    def StrDateDifference(self, sdate, edate):
        """ display qualitative difference between two dates """
        L = self.getL()
        if sdate > edate:
            raise "InvalidDates", L['INVALID_DATES']
        difference = EconoUtils.timeSince(sdate, edate, use_oldword=0, L=L)
        if type(difference)==type('s'):
            return difference.strip()
        else:
            return difference

    security.declareProtected('View', 'DayDifference')
    def DayDifference(self, sdate, edate):
        """ return difference in amount of days """
        if type(sdate)==type('s'):
            sdate = DateTime("%s 00:01"%sdate)
        if type(edate)==type('s'):
            edate = DateTime("%s 00:01"%edate)
        date_difference = int(round(edate - sdate))

        # for statistics subtract the days from timeskips
        date_difference = date_difference - self.TotalTimeskipDuration()

        return date_difference

    security.declareProtected('View', 'TotalSpent')
    def TotalSpent(self, startdate=None, enddate=None,
                   spendcodes=[], userids=[]):
        """ return Total amount per dates and spendcodes """
        # sort out input params
        if startdate and type(startdate)==type('s'):
            if startdate.strip()=='':
                startdate = None
            else:
                startdate = DateTime(startdate)
        if enddate and type(enddate)==type('s'):
            if enddate.strip()=='':
                enddate = None
            else:
                enddate = DateTime(enddate)
        # get ALL expenditures with spendcodes from 'spendcodes'
        expenditures = self.getExpenditures(spendcodes, userids)
        # sort them by 'datum' attribute
        expenditures = sequence.sort(expenditures, (('datum',),))
        # subtotal
        t=0
        for expenditure in expenditures:
            add = 1
            if startdate and expenditure.datum < startdate:
                add = 0
            if enddate and expenditure.datum > enddate:
                add = 0
            if add:
                t = t + expenditure.amount
        
        return t

    def getShowonlySpendcodeIds(self):
        """ for the REQUEST object fix up which ids to use """
        request = self.REQUEST
        ids = request.get('showonly_spendcodes',[])
        if 'ALL' in ids:
            ids = []
        return ids

    # ---------- FOR USERS ----------------------------------------

    def userExists(self, id):
        """ check if a user exists """
        if hasattr(self.users, id):
            return true
        else:
            return false
        
    def manage_addUser(self):
        """ add User via the web """
        L = self.getL()
        request = self.REQUEST
        for each in ['firstname','lastname','email','acl_username']:
            request.set(each, request.get(each,'').strip() )
        if request.firstname == '' and request.lastname == '':
            raise "InsufficientInput", L['SPECIFY_NAME_MORE']
            return 
        else:
            id = self.addUser(request.firstname,
                              request.lastname,
                              request.email,
                              request.acl_user,
                              return_id=1)

            url = '%s?manage_tabs_message=%s&r=%s'%\
                  (self.getRoot().absolute_url()+'/manage_users',
                   url_quote_plus(L['USER_ADDED']),
                   str(int(DateTime())))
            request.RESPONSE.redirect(url)
        
            
    def addUser(self, firstname, lastname,
                email=None, acl_username=None,
                REQUEST=None, return_id=0):
        """ create Econo User object """
        L = self.getL()
        if not email:
            email = ''
        if not acl_username:
            acl_username = ''
        firstname = firstname.strip()
        lastname = lastname.strip()
        email = email.strip()
        acl_username = acl_username.strip()
        id = self._generateUserId(firstname, lastname)
        user = EconoUser(id, firstname, lastname, email, acl_username)
        self.users._setObject(id, user)

        if return_id:
            return id
        elif REQUEST is not None:
            return self.manage_users(manage_tabs_message=L['USER_ADDED'])
        else:
            return id
        

    def _generateUserId(self, firstname, lastname):
        """ generate an suitable object id for a user object """
        id = str(int(DateTime()))
        if firstname != '' and lastname !='':
            id = "%s_%s.%s" % (id, firstname, lastname)
        elif firstname == '':
            id = "%s_%s" % (id, lastname)
        elif lastname == '':
            id = "%s_%s" % (id, firstname)
        id = string.translate(id, TRANSTAB_TABLE)
        id = EconoUtils.safeId(id)
        id = id[:40]
        return id
        
    def removeUsers(self, userids=[]):
        """ remove some objects simply """
        L = self.getL()
        nr_users = len(userids)
        if userids:
            if type(userids) == type('s'):
                userids = [userids]
            self.users.manage_delObjects(userids)
        # how many now?
        if len(self.getUsers()) <= 1:
            # turn of use_users
            self.UseUsers(0)
        # prevent silly reload problems
        url = "%s?manage_tabs_message=%s&r=%s"%\
              (self.getRoot().absolute_url()+ '/manage_users',
               url_quote_plus(L['NR_USERS_REMOVED']%nr_users),
               str(int(DateTime())))

        self.REQUEST.RESPONSE.redirect(url)
               

    security.declareProtected('Change properties', 'manage_useusers')                              
    def manage_useusers(self, REQUEST=None):
        """ change the use_users attribute """
        L = self.getL()
        action ="" 
        if self.UseUsers():
            # turn it off
            self.UseUsers(0)
            action = L['ACTION_USERS_OFF']
        else:
            if len(self.getUsers()) >1:
                self.UseUsers(1)
                action = L['ACTION_USERS_ON']
            else:
                action = L['ACTION_TOO_FEW_USERS']
        if REQUEST is not None:
            url = "%s?manage_tabs_message=%s&r=%s"%\
                  (self.getRoot().absolute_url()+'/manage_users',
                   action,
                   int(DateTime()))
            self.REQUEST.RESPONSE.redirect(url)

    security.declareProtected('View', 'UseUsers')
    def UseUsers(self, change=None):
        """ see if the property is on """
        if change is not None:
            if change:
                self.use_users=true
            else:
                self.use_users=false
        else:
            if self.use_users:
                return true
            else:
                return false
            
    security.declareProtected('View','getUsers')
    def getUsers(self, sort=0):
        """ return all econo user objects """
        us = self.users.objectValues('Econo User')
        if sort and self.userid_order:
            uss=[]
            for each in self.userid_order:
                uss.append(getattr(self.user,each))
            for each in us:
                if each not in uss:
                    uss.append(each)
            return uss            
        else:
            return us

    security.declareProtected('View','getUser')
    def getUser(self, id):
        """ get user object from its id """
        try:
            return getattr(self.users, id)
        except:
            return None
            

    security.declareProtected('View','suggestThisUser')
    def suggestThisUser(self, all_users=None):
        """ suggest who you are """
        request = self.REQUEST
        if all_users is None:
            all_users = self.getUsers()
        auth_user = self.getAuthenticUser(all_users)
        if auth_user:
            return auth_user.id
        elif self.use_cookies and request.has_key('ea_userid') and \
             hasattr(self.users, request.get('ea_userid')):            
            return getattr(self.users, request.get('ea_userid')).id
        else:
            return ''
            
        
    def getAuthenticUser(self, all_users=None):
        """ return Econo User object if this authentic user exists """
        if all_users is None:
            all_users = self.getUsers()
        auth_user = self.REQUEST.AUTHENTICATED_USER.getUserName()
        for user in all_users:
            if user.acl_username == auth_user:
                return user
        return None

    security.declareProtected('View', 'getUserFormSize')
    def getUserFormSize(self, nr_users):
        """ for the form """
        # use same logic as for spendcodes
        return self.getSpendcodeFormSize(nr_users)

    def getShowonlyUserIds(self):
        """ for the REQUEST object fix up which ids to use """
        request = self.REQUEST
        ids = request.get('showonly_userids',[])
        if 'ALL' in ids:
            ids = []
        return ids
    
    security.declareProtected('View management screens','manage_users')
    manage_users = DTMLFile('manage_users', globals())
                
InitializeClass(EconoAccount)


# # # # # # # # # # #
#
#  EconoExpenditure
#
#

class EconoExpenditure(SimpleItem.SimpleItem, EconoAccount):
    """ Econo Expenditure """
    meta_type=EXPENDITURE_META_TYPE
    meta_types=[]

    icon = "misc_/EconoAccount/expenditureicon.gif"

    manage_options = (
        {'label': 'Manage expenditure', 'action': 'manage_expenditure'},
        )
    
    _properties=({'id':'amount',    'type':'float',  'mode':'w'},
                 {'id':'datum',     'type':'date',   'mode':'w'},
                 {'id':'spendcode', 'type':'string', 'mode':'w'},
                 {'id':'userid',    'type':'string', 'mode':'w'},
                 {'id':'description','type':'text', 'mode':'w'}
                 )

    security=ClassSecurityInfo()
    
    def __init__(self, id, amount, datum, spendcode,
                 userid='', description=''):
        " "
        self.id = id
        self.amount = float(amount)
        if type(datum)==type('s'):
            datum = DateTime(datum)
        self.datum = datum
        self.spendcode = spendcode
        self.userid = userid
        self.description = description

    security.declareProtected('Manage properties','editExpenditure')
    def editExpenditure(self, amount=None, spendcode=None, datum=None,
                        userid=None, description=None):
        """ crude wrapper around manage_editExpenditure() """
        request = self.REQUEST
        
        if amount is not None:      request.set('amount',amount)
        if spendcode is not None:   request.set('spendcode',spendcode)
        if datum is not None:       request.set('datum',datum)
        if userid is not None:      request.set('userid', userid)
        if description is not None: request.set('description', description)
        return self.manage_editExpenditure()

    security.declareProtected('Manage properties','manage_editExpenditure')
    def manage_editExpenditure(self, returnmethodid=None, REQUEST=None):
        """ edit an expenditure via the web """
        request = self.REQUEST
        L = self.getL()
        if request.has_key('amount'):
            # check amount then change
            try:
                a=self._tuneAmount(request.get('amount'))
                self.amount = a
            except:
                pass
        if request.has_key('spendcode'):
            # check spendcode then change
            spendcode = request.get('spendcode').strip()
            if self._spendcodeExists(spendcode):
                self.spendcode = spendcode
        if request.has_key('datum'):
            # check datum then change
            datum = request.get('datum').strip()
            passed_date_test=0
            if self._IsValidDateString(datum):
                try:
                    datum = DateTime(datum)
                    passed_date_test=1
                except:
                    pass                       
            if passed_date_test:
                self.datum = datum
        if request.has_key('userid'):
            if self.userExists(request.get('userid')):
                self.userid = request.get('userid')        

        if returnmethodid:
            tabs_message = L['EXPENDITURE_CHANGED']
            return getattr(self, returnmethodid)(manage_tabs_message=tabs_message)
        elif REQUEST is not None:
            # to prevent silly reloads<->re-add
            url = '%s?manage_tabs_message=%s&r=%s'%\
                  (self.absolute_url(),
                   url_quote_plus(L['EXPENDITURE_CHANGED']),
                   str(int(DateTime())))
            request.RESPONSE.redirect(url)

    def getDescription(self, format="plaintext"):
        """ return description """
        format = format.strip().lower().replace('_','').replace('-','')                                                
        if format =="structuredtext":
            return structured_text(self.description)
        elif format =="html":
            return html_quote(self.description)
        else:
            return self.description

    def getAmount(self):
        """ return property """
        return self.amount

    def getDate(self):
        """ return property """
        return self.datum

    def getSpendcode(self):
        """ return property """
        return self.spendcode

    def getUserid(self):
        """ return property """
        return self.userid

    def getDescription(self):
        """ return property """
        return self.description

    security.declareProtected('View','getSpendcodeDescription')
    def getSpendcodeDescription(self, spendcode=None, only_title=0):
        """ return spendcode title and description in special format """
        if spendcode is None:
            spendcode = self.spendcode
        # spendcode object
        so = getattr(self.spendcodes,spendcode)
        if only_title:
            return so.getTitle()
        else:
            if so.has_description():
                return "%s (%s)"%(so.getTitle(), so.getDescription())
            else:
                return so.getTitle()

    security.declareProtected('View', 'index_html')
    index_html = DTMLFile('expenditure_index', globals())

    security.declareProtected('View', 'expenditure_information')
    expenditure_information = DTMLFile('expenditure_information', globals())

    security.declareProtected('Access contents information', 'manage_expenditure')
    manage_expenditure = DTMLFile('manage_expenditure', globals())

    security.declareProtected('View','remove_expenditure_form')
    remove_expenditure_form = DTMLFile('remove_expenditure_form', globals())

InitializeClass(EconoExpenditure)



# # # # # # # # # # #
#
#  EconoSpendcode
#
#

class EconoSpendcode(SimpleItem.SimpleItem, EconoAccount):
    """ Simple object """
    meta_type=SPENDCODE_META_TYPE
    meta_types=[]

    icon = "misc_/EconoAccount/spendcodeicon.gif"
    manage_options = (
        {'label': 'Manage spendcode', 'action': 'manage_spendcode'},
        )
    
    _properties=({'id':'title', 'type': 'string', 'mode':'w'},
                 {'id':'description', 'type': 'text', 'mode':'w'}
                 )

    security=ClassSecurityInfo()
    
    def __init__(self, id, title, description=''):
        """ important docstring :) """
        self.id = id
        self.title = title
        self.description = description

    security.declareProtected('View','index_html')
    index_html = DTMLFile('spendcode_index', globals())

    security.declareProtected('View','remove_spendcode_form')
    remove_spendcode_form = DTMLFile('remove_spendcode_form', globals())

    security.declareProtected('Access contents information','has_description')
    def has_description(self):
        " is description property set? "
        if self.description and self.description.strip() != '':
            return 1
        else:
            return 0

    security.declareProtected('Access contents information','getTitle')
    def getTitle(self):
        " return only title "
        return self.title

    security.declareProtected('Access contents information','getDescription')
    def getDescription(self):
        " return only description "
        if self.has_description():
            return self.description
        else:
            return ''

    security.declareProtected('Manage properties','manage_editSpendcode')
    def manage_editSpendcode(self):
        """ take out data from the REQUEST and save """
        L = self.getL()
        request = self.REQUEST
        title = request.get('title', self.title)
        description = request.get('description', self.description)
        self.editSpendcode(title, description)
        tabs_message = L['SPENDCODE_CHANGED']
        return self.manage_spendcode(manage_tabs_message=tabs_message)
    
    security.declareProtected('Manage properties','editSpendcode')
    def editSpendcode(self, title, description):
        """ change the spendcode """
        L = self.getL()
        self.title = title.strip()
        self.description = description.strip()
        tabs_message = L['SPENDCODE_CHANGED']
        return self.index_html(manage_tabs_message=tabs_message)


    security.declareProtected('Access contents information','spendcode_information')
    spendcode_information = DTMLFile('spendcode_information', globals())

    security.declareProtected('Access contents information','manage_spendcode')   
    manage_spendcode = DTMLFile('manage_spendcode', globals())

InitializeClass(EconoSpendcode)



# # # # # # # # # # #
#
#  EconoUser
#
#

class EconoUser(SimpleItem.SimpleItem, EconoAccount):
    """ Econo User """
    meta_type='Econo User'
    meta_types=[]

    icon = "misc_/EconoAccount/usericon.gif"

    manage_options = (
        {'label': 'Manage user', 'action': 'manage_user'},
        )
    
    _properties=({'id':'firstname', 'type':'string', 'mode':'w'},
                 {'id':'lastname',  'type':'string', 'mode':'w'},
                 {'id':'email',     'type':'string', 'mode':'w'},
                 {'id':'acl_username', 'type':'string', 'mode':'w'},
                 )

    security=ClassSecurityInfo()
    
    def __init__(self, id, firstname, lastname,
                 email='', acl_username=''):
        """ init EconoUser class """
        self.id = id
        self.firstname = firstname.strip()
        self.lastname = lastname.strip()
        self.email = email.strip()
        self.acl_username = acl_username.strip()

    def getFirstname(self):
        """ return firstname """
        return self.firstname

    def getLastname(self):
        """ return lastname """
        return self.lastname

    def getEmail(self):
        """ return email """
        return self.email

    def getACLUsername(self):
        """ return acl_username """
        return self.acl_username

    def getName(self, capitalize=0):
        """ return first and last name together """
        L = self.getL()
        if self.firstname != '' and self.lastname != '':
            name = self.firstname + " " + self.lastname
        elif self.firstname !='':
            name = self.firstname
        elif lastname !='':
            name = self.lastname
        elif email != '':
            name = self.email
        else:
            raise "IncompleteUser", L['USER_INSUFFICIENT_INFORMATION']
        if capitalize:
            return name.capitalize()
        else:
            return name

    def ValidEmailAddress(self):
        """ check if the email address given is valid """
        if EconoUtils.ValidEmailAddress(self.email):
            return true
        else:
            return false

    security.declareProtected('View management screens', 'manage_user')
    manage_user = DTMLFile('manage_user', globals())
        
InitializeClass(EconoUser)        



# # # # # # # # # # #
#
#  EconoTimeSkip
#
#

class EconoTimeskip(SimpleItem.SimpleItem, EconoAccount):
    """ Econo TimeSkip """
    meta_type= TIMESKIP_META_TYPE
    meta_types=[]

    icon = "misc_/EconoAccount/timeskipicon.gif"

    manage_options = (
        {'label': 'Timeskips', 'action': 'manage_timeskips'},
        )
    
    _properties=({'id':'start_date', 'type':'date', 'mode':'w'},
                 {'id':'end_date',   'type':'date', 'mode':'w'},
                 {'id':'description','type':'text', 'mode':'w'},
                 )

    security=ClassSecurityInfo()
    
    def __init__(self, id, start_date, end_date, description=''):
        """ init EconoTimeSkip class """
        self.id = id
        self.start_date = self._datetype_assurance(start_date)
        self.end_date = self._datetype_assurance(end_date)
        self.description = description

    def _datetype_assurance(self, d):
        """ make sure a date is a date """
        if type(d) == type('s'):
            d = d.strip()
            d = DateTime(d)
        return d

    def getId(self):
        """ return id """
        return self.id

    def getStartdate(self):
        """ return start_date """
        return self.start_date

    def getEnddate(self):
        """ return end_date """
        return self.end_date

    def getDescription(self):
        """ return description """
        return self.description

    security.declareProtected('View', 'getDuration')
    def getDuration(self, returntype="float"):
        """ return duration between start and end date"""
        L = self.getL()
        returntype = returntype.lower().strip()
        if returntype == "string":
            return self.StrDateDifference(self.start_date, self.end_date)
        elif returntype == "int":
            return int(self.end_date - self.start_date)
        elif returntype == "float":
            return float(self.end_date - self.start_date)
        else:
            raise "UnknownReturntype", L['UNKNOWN_RETURNTYPE']

    def DateSpanned(self, date):
        """ Check if this object spans this 'date' """
        if type(date) == type('s'):
            date = DateTime(date)

        if date > self.start_date and date < self.end_date:
            return 1
        else:
            return 0
        
        

        
InitializeClass(EconoTimeskip)
